;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "fussy" "2.2"
  "Fuzzy completion style using `flx'."
  '((emacs  "28.2")
    (flx    "0.5")
    (compat "30.0.0.0"))
  :url "https://github.com/jojojames/fussy"
  :commit "92fb91c034707af77be9869500a5ae1ea0079b7d"
  :revdesc "92fb91c03470"
  :keywords '("matching")
  :authors '(("James Nguyen" . "james@jojojames.com"))
  :maintainers '(("James Nguyen" . "james@jojojames.com")))
